from django_prometheus.db import (
    connections_total, execute_total, execute_many_total, errors_total,
    connection_errors_total)


class ExceptionCounterByType(object):
    """A context manager that counts exceptions by type.

    Exceptions increment the provided counter, whose last label's name
    must match the `type_label` argument.

    In other words:

    c = Counter('http_request_exceptions_total', 'Counter of exceptions',
                ['method', 'type'])
    with ExceptionCounterByType(c, extra_labels={'method': 'GET'}):
        handle_get_request()
    """

    def __init__(self, counter, type_label='type', extra_labels=None):
        self._counter = counter
        self._type_label = type_label
        self._labels = extra_labels

    def __enter__(self):
        pass

    def __exit__(self, typ, value, traceback):
        if typ is not None:
            self._labels.update({self._type_label: typ.__name__})
            self._counter.labels(**self._labels).inc()


class DatabaseWrapperMixin(object):
    """Extends the DatabaseWrapper to count connections and cursors."""

    def get_new_connection(self, *args, **kwargs):
        connections_total.labels(self.alias, self.vendor).inc()
        try:
            return super(DatabaseWrapperMixin, self).get_new_connection(
                *args, **kwargs)
        except:
            connection_errors_total.labels(self.alias, self.vendor).inc()
            raise

    def create_cursor(self, name=None):
        return self.connection.cursor(factory=ExportingCursorWrapper(
            self.CURSOR_CLASS, self.alias, self.vendor))


def ExportingCursorWrapper(cursor_class, alias, vendor):
    """Returns a CursorWrapper class that knows its database's alias and
    vendor name.
    """

    class CursorWrapper(cursor_class):
        """Extends the base CursorWrapper to count events."""

        def execute(self, *args, **kwargs):
            execute_total.labels(alias, vendor).inc()
            with ExceptionCounterByType(errors_total, extra_labels={
                    'alias': alias, 'vendor': vendor}):
                return super(CursorWrapper, self).execute(*args, **kwargs)

        def executemany(self, query, param_list, *args, **kwargs):
            execute_total.labels(alias, vendor).inc(len(param_list))
            execute_many_total.labels(alias, vendor).inc(len(param_list))
            with ExceptionCounterByType(errors_total, extra_labels={
                    'alias': alias, 'vendor': vendor}):
                return super(CursorWrapper, self).executemany(
                    query, param_list, *args, **kwargs)
    return CursorWrapper



import copy
from prometheus_client import REGISTRY


METRIC_EQUALS_ERR_EXPLANATION = """
%s%s = %s, expected %s.
The values for %s are:
%s"""

METRIC_DIFF_ERR_EXPLANATION = """
%s%s changed by %f, expected %f.
Value before: %s
Value after: %s
"""

METRIC_COMPARE_ERR_EXPLANATION = """
The change in value of %s%s didn't match the predicate.
Value before: %s
Value after: %s
"""

METRIC_DIFF_ERR_NONE_EXPLANATION = """
%s%s was None after.
Value before: %s
Value after: %s
"""


class PrometheusTestCaseMixin(object):
    """A collection of utilities that make it easier to write test cases
    that interact with metrics.
    """

    def saveRegistry(self, registry=REGISTRY):
        """Freezes a registry. This lets a user test changes to a metric
        instead of testing the absolute value. A typical use case looks like:

          r = self.saveRegistry()
          doStuff()
          self.assertMetricDiff(r, 1, 'stuff_done_total')
        """
        return copy.deepcopy(list(registry.collect()))

    def getMetricFromFrozenRegistry(self, metric_name, frozen_registry,
                                    **labels):
        """Gets a single metric from a frozen registry."""
        for metric in frozen_registry:
            for n, l, value in metric.samples:
                if n == metric_name and l == labels:
                    return value

    def getMetric(self, metric_name, registry=REGISTRY, **labels):
        """Gets a single metric."""
        return self.getMetricFromFrozenRegistry(
            metric_name, registry.collect(), **labels)

    def getMetricVectorFromFrozenRegistry(self, metric_name, frozen_registry):
        """Like getMetricVector, but from a frozen registry."""
        output = []
        for metric in frozen_registry:
            for n, l, value in metric.samples:
                if n == metric_name:
                    output.append((l, value))
        return output

    def getMetricVector(self, metric_name, registry=REGISTRY):
        """Returns the values for all labels of a given metric.

        The result is returned as a list of (labels, value) tuples,
        where `labels` is a dict.

        This is quite a hack since it relies on the internal
        representation of the prometheus_client, and it should
        probably be provided as a function there instead.
        """
        return self.getMetricVectorFromFrozenRegistry(
            metric_name, registry.collect())

    def formatLabels(self, labels):
        """Format a set of labels to Prometheus representation.

        In:
          {'method': 'GET', 'port': '80'}

        Out:
          '{method="GET",port="80"}'
        """
        return '{%s}' % ','.join([
            '%s="%s"' % (k, v) for k, v in labels.items()])

    def formatVector(self, vector):
        """Formats a list of (labels, value) where labels is a dict into a
        human-readable representation.
        """
        return '\n'.join([
            '%s = %s' % (self.formatLabels(labels), value)
            for labels, value in vector])

    def assertMetricEquals(self, expected_value, metric_name,
                           registry=REGISTRY, **labels):
        """Asserts that metric_name{**labels} == expected_value."""
        value = self.getMetric(metric_name, registry=registry, **labels)
        self.assertEqual(
            expected_value, value, METRIC_EQUALS_ERR_EXPLANATION % (
                metric_name, self.formatLabels(labels), value,
                expected_value, metric_name,
                self.formatVector(self.getMetricVector(metric_name))))

    def assertMetricDiff(self, frozen_registry, expected_diff,
                         metric_name, registry=REGISTRY, **labels):
        """Asserts that metric_name{**labels} changed by expected_diff between
        the frozen registry and now. A frozen registry can be obtained
        by calling saveRegistry, typically at the beginning of a test
        case.
        """
        saved_value = self.getMetricFromFrozenRegistry(
            metric_name, frozen_registry, **labels)
        current_value = self.getMetric(metric_name, registry=registry,
                                       **labels)
        self.assertFalse(
            current_value is None,
            METRIC_DIFF_ERR_NONE_EXPLANATION % (
                metric_name, self.formatLabels(labels),
                saved_value,
                current_value))
        diff = current_value - (saved_value or 0.0)
        self.assertEqual(
            expected_diff, diff,
            METRIC_DIFF_ERR_EXPLANATION % (
                metric_name, self.formatLabels(labels), diff, expected_diff,
                saved_value,
                current_value))

    def assertMetricCompare(self, frozen_registry, predicate,
                            metric_name, registry=REGISTRY, **labels):
        """Asserts that metric_name{**labels} changed according to a provided
        predicate function between the frozen registry and now. A
        frozen registry can be obtained by calling saveRegistry,
        typically at the beginning of a test case.
        """
        saved_value = self.getMetricFromFrozenRegistry(
            metric_name, frozen_registry, **labels)
        current_value = self.getMetric(metric_name, registry=registry,
                                       **labels)
        self.assertFalse(
            current_value is None,
            METRIC_DIFF_ERR_NONE_EXPLANATION % (
                metric_name, self.formatLabels(labels),
                saved_value,
                current_value))
        self.assertTrue(
            predicate(saved_value, current_value),
            METRIC_COMPARE_ERR_EXPLANATION % (
                metric_name, self.formatLabels(labels),
                saved_value,
                current_value))



"""Django-Prometheus

https://github.com/korfuri/django-prometheus
"""

# Import all files that define metrics. This has the effect that
# `import django_prometheus` will always instanciate all metric
# objects right away.
import django_prometheus.middleware
import django_prometheus.models

# Import pip_prometheus to export the pip metrics automatically.
try:
    import pip_prometheus
except ImportError:
    # If people don't have pip, don't export anything.
    pass

default_app_config = 'django_prometheus.apps.DjangoPrometheusConfig'



from prometheus_client import Counter

model_inserts = Counter(
    'django_model_inserts_total',
    'Number of insert operations by model.',
    ['model'])

model_updates = Counter(
    'django_model_updates_total',
    'Number of update operations by model.',
    ['model'])

model_deletes = Counter(
    'django_model_deletes_total',
    'Number of delete operations by model.',
    ['model'])


def ExportModelOperationsMixin(model_name):
    """Returns a mixin for models to export counters for lifecycle operations.

    Usage:
      class User(ExportModelOperationsMixin('user'), Model):
          ...
    """
    # Force create the labels for this model in the counters. This
    # is not necessary but it avoids gaps in the aggregated data.
    model_inserts.labels(model_name)
    model_updates.labels(model_name)
    model_deletes.labels(model_name)

    class Mixin(object):
        def _do_insert(self, *args, **kwargs):
            model_inserts.labels(model_name).inc()
            return super(Mixin, self)._do_insert(*args, **kwargs)

        def _do_update(self, *args, **kwargs):
            model_updates.labels(model_name).inc()
            return super(Mixin, self)._do_update(*args, **kwargs)

        def delete(self, *args, **kwargs):
            model_deletes.labels(model_name).inc()
            return super(Mixin, self).delete(*args, **kwargs)
    return Mixin



from timeit import default_timer


def Time():
    """Returns some representation of the current time.

    This wrapper is meant to take advantage of a higher time
    resolution when available. Thus, its return value should be
    treated as an opaque object. It can be compared to the current
    time with TimeSince().
    """
    return default_timer()


def TimeSince(t):
    """Compares a value returned by Time() to the current time.

    Returns:
      the time since t, in fractional seconds.
    """
    return default_timer() - t


def PowersOf(logbase, count, lower=0, include_zero=True):
    """Returns a list of count powers of logbase (from logbase**lower)."""
    if not include_zero:
        return [logbase ** i for i in range(lower, count+lower)]
    else:
        return [0] + [logbase ** i for i in range(lower, count+lower)]



from django.http import HttpResponse
from django.conf import settings
from prometheus_client import multiprocess

try:
    # Python 2
    from BaseHTTPServer import HTTPServer
except ImportError:
    # Python 3
    from http.server import HTTPServer
import socket
import logging
import os
import prometheus_client
import threading


logger = logging.getLogger(__name__)


def SetupPrometheusEndpointOnPort(port, addr=''):
    """Exports Prometheus metrics on an HTTPServer running in its own thread.

    The server runs on the given port and is by default listenning on
    all interfaces. This HTTPServer is fully independent of Django and
    its stack. This offers the advantage that even if Django becomes
    unable to respond, the HTTPServer will continue to function and
    export metrics. However, this also means that the features
    offered by Django (like middlewares or WSGI) can't be used.

    Now here's the really weird part. When Django runs with the
    auto-reloader enabled (which is the default, you can disable it
    with `manage.py runserver --noreload`), it forks and executes
    manage.py twice. That's wasteful but usually OK. It starts being a
    problem when you try to open a port, like we do. We can detect
    that we're running under an autoreloader through the presence of
    the RUN_MAIN environment variable, so we abort if we're trying to
    export under an autoreloader and trying to open a port.
    """
    assert os.environ.get('RUN_MAIN') != 'true', (
        'The thread-based exporter can\'t be safely used when django\'s '
        'autoreloader is active. Use the URL exporter, or start django '
        'with --noreload. See documentation/exports.md.')
    prometheus_client.start_http_server(port, addr=addr)


class PrometheusEndpointServer(threading.Thread):
    """A thread class that holds an http and makes it serve_forever()."""
    def __init__(self, httpd, *args, **kwargs):
        self.httpd = httpd
        super(PrometheusEndpointServer, self).__init__(*args, **kwargs)

    def run(self):
        self.httpd.serve_forever()


def SetupPrometheusEndpointOnPortRange(port_range, addr=''):
    """Like SetupPrometheusEndpointOnPort, but tries several ports.

    This is useful when you're running Django as a WSGI application
    with multiple processes and you want Prometheus to discover all
    workers. Each worker will grab a port and you can use Prometheus
    to aggregate across workers.

    port_range may be any iterable object that contains a list of
    ports. Typically this would be an xrange of contiguous ports.

    As soon as one port is found that can serve, use this one and stop
    trying.

    The same caveats regarding autoreload apply. Do not use this when
    Django's autoreloader is active.

    """
    assert os.environ.get('RUN_MAIN') != 'true', (
        'The thread-based exporter can\'t be safely used when django\'s '
        'autoreloader is active. Use the URL exporter, or start django '
        'with --noreload. See documentation/exports.md.')
    for port in port_range:
        try:
            httpd = HTTPServer((addr, port), prometheus_client.MetricsHandler)
        except (OSError, socket.error):
            # Python 2 raises socket.error, in Python 3 socket.error is an
            # alias for OSError
            continue  # Try next port
        thread = PrometheusEndpointServer(httpd)
        thread.daemon = True
        thread.start()
        logger.info('Exporting Prometheus /metrics/ on port %s' % port)
        return  # Stop trying ports at this point


def SetupPrometheusExportsFromConfig():
    """Exports metrics so Prometheus can collect them."""
    port = getattr(settings, 'PROMETHEUS_METRICS_EXPORT_PORT', None)
    port_range = getattr(
        settings, 'PROMETHEUS_METRICS_EXPORT_PORT_RANGE', None)
    addr = getattr(settings, 'PROMETHEUS_METRICS_EXPORT_ADDRESS', '')
    if port_range:
        SetupPrometheusEndpointOnPortRange(port_range, addr)
    elif port:
        SetupPrometheusEndpointOnPort(port, addr)


def ExportToDjangoView(request):
    """Exports /metrics as a Django view.

    You can use django_prometheus.urls to map /metrics to this view.
    """
    if 'prometheus_multiproc_dir' in os.environ:
        registry = prometheus_client.CollectorRegistry()
        multiprocess.MultiProcessCollector(registry)
    else:
        registry = prometheus_client.REGISTRY
    metrics_page = prometheus_client.generate_latest(registry)
    return HttpResponse(
        metrics_page,
        content_type=prometheus_client.CONTENT_TYPE_LATEST)



#!/usr/bin/env python
from django_prometheus.testutils import PrometheusTestCaseMixin
import prometheus_client
from operator import itemgetter
import unittest


class SomeTestCase(PrometheusTestCaseMixin):
    """A class that pretends to be a unit test."""

    def __init__(self):
        self.passes = True
        super(SomeTestCase, self).__init__()

    def assertEqual(self, l, r, *args, **kwargs):
        self.passes = self.passes and (l == r)

    def assertFalse(self, e, *args, **kwargs):
        self.passes = self.passes and (not e)


class PrometheusTestCaseMixinTest(unittest.TestCase):

    def setUp(self):
        self.registry = prometheus_client.CollectorRegistry()
        self.some_gauge = prometheus_client.Gauge(
            'some_gauge', 'Some gauge.', registry=self.registry)
        self.some_gauge.set(42)
        self.some_labelled_gauge = prometheus_client.Gauge(
            'some_labelled_gauge', 'Some labelled gauge.', [
                'labelred', 'labelblue'], registry=self.registry)
        self.some_labelled_gauge.labels('pink', 'indigo').set(1)
        self.some_labelled_gauge.labels('pink', 'royal').set(2)
        self.some_labelled_gauge.labels('carmin', 'indigo').set(3)
        self.some_labelled_gauge.labels('carmin', 'royal').set(4)
        self.t = SomeTestCase()

    def testGetMetrics(self):
        """Tests getMetric."""
        self.assertEqual(42, self.t.getMetric(
            'some_gauge', registry=self.registry))
        self.assertEqual(
            1,
            self.t.getMetric(
                'some_labelled_gauge',
                registry=self.registry,
                labelred='pink',
                labelblue='indigo'))

    def testGetMetricVector(self):
        """Tests getMetricVector."""
        vector = self.t.getMetricVector(
            'some_nonexistent_gauge', registry=self.registry)
        self.assertEqual([], vector)
        vector = self.t.getMetricVector('some_gauge', registry=self.registry)
        self.assertEqual([({}, 42)], vector)
        vector = self.t.getMetricVector(
            'some_labelled_gauge', registry=self.registry)
        self.assertEqual(sorted([
            ({'labelred': u'pink', 'labelblue': u'indigo'}, 1),
            ({'labelred': u'pink', 'labelblue': u'royal'}, 2),
            ({'labelred': u'carmin', 'labelblue': u'indigo'}, 3),
            ({'labelred': u'carmin', 'labelblue': u'royal'}, 4),
        ], key=itemgetter(1)), sorted(vector, key=itemgetter(1)))

    def testAssertMetricEquals(self):
        """Tests assertMetricEquals."""
        # First we test that a scalar metric can be tested.
        self.t.assertMetricEquals(42, 'some_gauge', registry=self.registry)
        self.assertTrue(self.t.passes)
        self.t.assertMetricEquals(43, 'some_gauge', registry=self.registry)
        self.assertFalse(self.t.passes)
        self.t.passes = True

        # Here we test that assertMetricEquals fails on nonexistent gauges.
        self.t.assertMetricEquals(
            42, 'some_nonexistent_gauge', registry=self.registry)
        self.assertFalse(self.t.passes)
        self.t.passes = True

        # Here we test that labelled metrics can be tested.
        self.t.assertMetricEquals(
            1,
            'some_labelled_gauge',
            registry=self.registry,
            labelred='pink',
            labelblue='indigo')
        self.assertTrue(self.t.passes)
        self.t.assertMetricEquals(
            1,
            'some_labelled_gauge',
            registry=self.registry,
            labelred='tomato',
            labelblue='sky')
        self.assertFalse(self.t.passes)
        self.t.passes = True

    def testRegistrySaving(self):
        """Tests saveRegistry and frozen registries operations."""
        frozen_registry = self.t.saveRegistry(registry=self.registry)
        # Test that we can manipulate a frozen scalar metric.
        self.assertEqual(42, self.t.getMetricFromFrozenRegistry(
            'some_gauge', frozen_registry))
        self.some_gauge.set(99)
        self.assertEqual(42, self.t.getMetricFromFrozenRegistry(
            'some_gauge', frozen_registry))
        self.t.assertMetricDiff(frozen_registry, 99 -
                                42, 'some_gauge', registry=self.registry)
        self.assertTrue(self.t.passes)
        self.t.assertMetricDiff(
            frozen_registry, 1, 'some_gauge', registry=self.registry)
        self.assertFalse(self.t.passes)
        self.t.passes = True

        # Now test the same thing with a labelled metric.
        self.assertEqual(
            1,
            self.t.getMetricFromFrozenRegistry(
                'some_labelled_gauge',
                frozen_registry,
                labelred='pink',
                labelblue='indigo'))
        self.some_labelled_gauge.labels('pink', 'indigo').set(5)
        self.assertEqual(
            1,
            self.t.getMetricFromFrozenRegistry(
                'some_labelled_gauge',
                frozen_registry,
                labelred='pink',
                labelblue='indigo'))
        self.t.assertMetricDiff(
            frozen_registry,
            5 - 1,
            'some_labelled_gauge',
            registry=self.registry,
            labelred='pink',
            labelblue='indigo')
        self.assertTrue(self.t.passes)
        self.t.assertMetricDiff(
            frozen_registry,
            1,
            'some_labelled_gauge',
            registry=self.registry,
            labelred='pink',
            labelblue='indigo')
        self.assertFalse(self.t.passes)


if __name__ == 'main':
    unittest.main()
